#!/usr/bin/perl -w

use strict;
use FileHandle;
use File::Basename;
use lib(dirname($0));
use pattern5;

use constant STATE_INIT    => 0;
use constant STATE_LOGFILE => 1;

my $file = shift @ARGV;

my $fh = new FileHandle;
if ($file =~ /\.bz2$/) {
  $fh->open("bzcat $file |");
}
else {
  $fh->open("< $file");
}
die "Failed to open \"$file\" for reading\n" unless defined $fh;

my %lnkMap;

#-------------------------------------------------------------------------------
# Parse file
#-------------------------------------------------------------------------------

print "file = $file\n";

my $state  = STATE_INIT;
my $rx     = undef;
my $lnk    = undef;
my %epLst;
while (<$fh>) {
  if (/ssh.*root\@(10\.98\.\d+\.\d+)\s.*lnk(\d)/) {
    my $node = Node->new($1);
    my $phy  = $2;
    $rx      = Endpoint->new($node, $phy);
    $state   = STATE_LOGFILE;
    next;
  }

  if ($state == STATE_INIT && /^\s*(\d+):(\d+)\s+(\d)([pr])\s+(\d+):(\d+)\s+(\d)([pr])/) {
    my $a = Endpoint->new($1, $2, $3);
    my $b = Endpoint->new($5, $6, $7);
    $lnkMap{$a->as_string} = $b;
    $lnkMap{$b->as_string} = $a;
  }

  if ($state == STATE_INIT && /\s(\d+\.\d+\.\d+\.\d+)\s+lnk (\d) ([RT]X_.*)=(0x[0-9a-f]+)/) {
    my $node     = Node->new($1);
    my $endpoint = Endpoint->new($node, $2);
    my $reg      = $3;
    my $val      = hex($4);

    unless (defined $epLst{$endpoint->as_string}) {
      $epLst{$endpoint->as_string} = $endpoint;
    }
    $epLst{$endpoint->as_string}->{status}->{$reg} = $val;
  }

  if ($state == STATE_LOGFILE && /END/) {
    $state = STATE_INIT;
    next;
  }

  if ($state == STATE_LOGFILE && /^err\s.*0x/) {
    my @wLst;
    foreach (split /\s+/) {
      next unless ($_ =~ /^\s*0x/);
      s/0x//g;
      my @f = split /_/, $_;
      push @wLst, Pattern5->new(@f);
    }

    my $tx = $lnkMap{$rx->as_string};
    die unless defined $tx;

    printf ">>> line=%d\n    %s", $., $_;
    printf "    RX=%s TX=%s\n", $rx->as_string, $tx->as_string;
    printf "    > received\n";
    printf $wLst[0]->as_string("      ");
    printf "    > reference\n";
    printf $wLst[1]->as_string("      ");

    my @err = ana($tx, $rx, @wLst);

    if (scalar @err) {
      foreach (@err) {
        print "    --> $_\n";
      }
    }
    else {
      print "    --> UNKNOWN ERROR CASE\n";
    }

    $state = STATE_INIT;
  }
}


foreach my $key (sort keys %epLst) {
  my $ep = $epLst{$key};

  my $rxExc = $ep->{status}->{RX_EXC};
  if (defined $rxExc && ($rxExc & 0x37fb) != 0) {
    printf ">>> RX_EXC[%s] = 0x%08x (TX=%s)\n",
           $ep->as_string, $rxExc, $lnkMap{$ep->as_string}->as_string;
    printf "    RX_FAULT from PHY\n"             if ($rxExc & 0x00000003);
    printf "    Timeout\n"                       if ($rxExc & 0x00000008);
    printf "    Credit FIFO overrun\n"           if ($rxExc & 0x00000010);
    printf "    Credit size zero\n"              if ($rxExc & 0x00000020);
    printf "    rCba/rNba written while used\n"  if ($rxExc & 0x00000060);
    printf "    Data received while offline\n"   if ($rxExc & 0x00000100);
    printf "    OWC protocol error\n"            if ($rxExc & 0x00000600);
    printf "    Invalid rCba/rNba arithmetics\n" if ($rxExc & 0x00003000);
  }

  my $txExc = $ep->{status}->{TX_EXC};
  if (defined $txExc && ($txExc & 0x07f9) != 0) {
    printf ">>> TX_EXC[%s] = 0x%08x (RX=%s)\n",
           $ep->as_string, $txExc, $lnkMap{$ep->as_string}->as_string;
    printf "    TX_FAULT\n"                      if ($txExc & 0x00000001);
    printf "    Feedback mismatch\n"             if ($txExc & 0x00000008);
    printf "    Invalid header\n"                if ($txExc & 0x00000010);
    printf "    txFifo written while full\n"     if ($txExc & 0x00000020);
    printf "    txFifo empty during packet\n"    if ($txExc & 0x00000040);
    printf "    txBuffer full\n"                 if ($txExc & 0x00000080);
    printf "    Data written while offline\n"    if ($txExc & 0x00000100);
    printf "    fbIn accumulation\n"             if ($txExc & 0x00000200);
    printf "    fbOut accumulation\n"            if ($txExc & 0x00000400);
  }

  my $cntErr = $ep->{status}->{RX_DATA};
  if (defined $cntErr && $cntErr != 0) {
    printf ">>> RX_DATA[%s] = 0x%08x (TX=%s)\n",
           $ep->as_string, $cntErr, $lnkMap{$ep->as_string}->as_string;
    printf "    cntErrC = %d\n", ($cntErr & 0x0000ffff) >>  0 if ($cntErr & 0x0000ffff);
    printf "    cntErrD = %d\n", ($cntErr & 0x00ff0000) >> 16 if ($cntErr & 0x00ff0000);
    printf "    cntErrT = %d\n", ($cntErr & 0xff000000) >> 24 if ($cntErr & 0xff000000);
  }
}


#-------------------------------------------------------------------------------
# Identify error cases
#-------------------------------------------------------------------------------

sub ana
{
  my $tx   = $_[0];	# send link endpoint (type Endpoint)
  my $rx   = $_[1];	# receiver link endpoint (type Endpoint)
  my $rxW  = $_[2];	# decoded received word (type Pattern5)
  my $refW = $_[3];	# decoded reference word (type Pattern5)

  my @err;

  # Identify mismatches
  my $nodeMm = ($rxW->{tx}->{node}->hostname ne $tx->{node}->hostname) ?  1 : 0;
  my $linkMm = ($rxW->{tx}->as_string ne $tx->as_string) ? 1 : 0;
  my $chMm   = ($rxW->{ch} != $refW->{ch}) ? 1 : 0;
  my $iterMm = ($rxW->{iter} != $refW->{iter}) ? 1 : 0;
  my $iMsgMm = ($rxW->{iMsg} != $refW->{iMsg}) ? 1 : 0;
  my $a128Mm = ($rxW->a128 != $refW->a128) ? 1 : 0;

  my $a4Mm = 0;
  if ($rxW->na4 == 4) {
    for my $i (0..$rxW->na4-1) {
      $a4Mm = 1 if ($rxW->a4($i) != $i);
    }
  }

  # Identify inconsistencies in redundantly encoded information
  my $a128MmRed = 0;
  if ($rxW->na128 > 0) {
    for my $i (1..$rxW->na128-1) {
      $a128MmRed = 1 if ($rxW->a128($i) != $rxW->a128($i-1));
    }
  }

  my $a16MmRed = 0;
  if ($rxW->na16 > 0) {
    for my $i (1..$rxW->na16-1) {
      $a16MmRed = 1 if ($rxW->a16($i) != $rxW->a16($i-1));
    }
  }

  # Identify error cases
  if ($rxW->allZero()) {
    push @err, sprintf "lost packet";
  }
  elsif ($a4Mm) {
    push @err, sprintf "bad data: a4 counter does not match",
                       $rxW->{tx}->as_string;
  }
  elsif ($a128MmRed) {
    push @err, sprintf "bad data: 128 counter does not match",
                       $rxW->{tx}->as_string;
  }
  elsif ($a16MmRed) {
    push @err, sprintf "bad data: a16 counter does not match",
                       $rxW->{tx}->as_string;
  }
  elsif ($nodeMm) {
    push @err, sprintf "multihop: received data generated for %s",
                       $rxW->{tx}->as_string;
  }
  elsif ($linkMm) {
    push @err, sprintf "link/phy mismatch: expected %s, received %s",
                       $tx->as_string, $rxW->{tx}->as_string;
  }
  elsif ($chMm && !$iterMm && !$iMsgMm) {
    push @err, sprintf "channel mismatch: expected %d, received %d",
                       $refW->{ch}, $rxW->{ch};
  }
  elsif ($a128Mm && !$iterMm && !$iMsgMm) {
    push @err, sprintf "a128 mismatch: expect %d, received %d",
                       $refW->a128, $rxW->a128;
  }
  elsif ($iterMm) {
    push @err, sprintf "iteration counter mismatch: expected %d, received %d",
                       $refW->{iter}, $rxW->{iter};
  }

  return @err;
}
